#include "shiftjis_to_utf8.hpp"
#include "shiftjis.h"
#include <cstdint>
#if !defined(NDEBUG)
#	include <iostream>
#endif

namespace mc {
namespace {
	unsigned int sequence_length (std::string::const_iterator lead_it) {
		const uint8_t lead = static_cast<uint8_t>(*lead_it & 0xff);
		if (lead < 0x80)
			return 1;
		else if ((lead >> 5) == 0x6)
			return 2;
		else if ((lead >> 4) == 0xe)
			return 3;
		else if ((lead >> 3) == 0x1e)
			return 4;
		else
			return 0;
	}

	uint32_t to_codepoint (std::string::const_iterator it, unsigned int length) {
		uint32_t cp = static_cast<uint32_t>(*it);
		switch (length) {
		case 1:
			break;
		case 2:
			it++;
			cp = ((cp << 6) & 0x7ff) + (static_cast<uint8_t>(*it) & 0x3f);
			break;
		case 3:
			++it;
			cp = ((cp << 12) & 0xffff) + ((static_cast<uint8_t>(*it) << 6) & 0xfff);
			++it;
			cp += static_cast<uint8_t>(*it) & 0x3f;
			break;
		case 4:
			++it;
			cp = ((cp << 18) & 0x1fffff) + ((static_cast<uint8_t>(*it) << 12) & 0x3ffff);
			++it;
			cp += (static_cast<uint8_t>(*it) << 6) & 0xfff;
			++it;
			cp += static_cast<uint8_t>(*it) & 0x3f;
			break;
		}
		return cp;
	}

	//see https://stackoverflow.com/questions/15631168/validate-japanese-character-in-active-record-callback/15651264#15651264
	[[gnu::const]] bool is_fullwidth_roman (uint32_t cp) {
		return (cp >= 0xff01 and cp <= 0xff60);
	}
} //unnamed namespace

std::string shiftjis_to_utf8 (std::string_view in) {
	std::string output(3 * in.size(), ' '); //ShiftJis won't give 4byte UTF8, so max. 3 byte per input char are needed
	size_t indexInput = 0, indexOutput = 0;

	while(indexInput < in.size()) {
		char arraySection = ((uint8_t)in[indexInput]) >> 4;

		size_t arrayOffset;
		if(arraySection == 0x8) arrayOffset = 0x100; //these are two-byte shiftjis
		else if(arraySection == 0x9) arrayOffset = 0x1100;
		else if(arraySection == 0xE) arrayOffset = 0x2100;
		else arrayOffset = 0; //this is one byte shiftjis

		//determining real array offset
		if(arrayOffset) {
			arrayOffset += (((uint8_t)in[indexInput]) & 0xf) << 8;
			indexInput++;
			if(indexInput >= in.size()) break;
		}
		arrayOffset += (uint8_t)in[indexInput++];
		arrayOffset <<= 1;

		//unicode number is...
		uint16_t unicodeValue = (g_shiftjis[arrayOffset] << 8) | g_shiftjis[arrayOffset + 1];

		//converting to UTF8
		if(unicodeValue < 0x80) {
			output[indexOutput++] = unicodeValue;
		}
		else if(unicodeValue < 0x800) {
			output[indexOutput++] = 0xC0 | (unicodeValue >> 6);
			output[indexOutput++] = 0x80 | (unicodeValue & 0x3f);
		}
		else {
			output[indexOutput++] = 0xE0 | (unicodeValue >> 12);
			output[indexOutput++] = 0x80 | ((unicodeValue & 0xfff) >> 6);
			output[indexOutput++] = 0x80 | (unicodeValue & 0x3f);
		}
	}

	output.resize(indexOutput); //remove the unnecessary bytes
	return output;
}

std::string full_to_halfwidth_ascii (const std::string& in) {
	if (in.empty())
		return std::string();

	auto it = in.begin();
	std::string retval;
	do {
		const auto seq_len = sequence_length(it);
		const auto code = to_codepoint(it, seq_len);
		if (is_fullwidth_roman(code)) {
			retval.push_back(0xFF & (code + 0x20));
		}
		else if (12288 == code) {
			retval.push_back(' ');
		}
		else if (8722 == code) {
			retval.push_back('-');
		}
		else {
#if !defined(NDEBUG)
			std::cout << "unknown codepoint " << code << '\n';
#endif
			std::copy(it, it + seq_len, std::back_inserter(retval));
		}
		it += seq_len;
	} while (it != in.end());
	return retval;
}
} //namespace mc
